#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdarg.h>
#include <time.h>

#include "cra_c.h"

FILE *cra_c_debug = NULL;

char *strim_simple(char *s)
{
  char *s2;

  if (!s)
    return NULL;

  s2 = s + strlen(s)-1;

  while (*s2 <= 32 && *s2 > 0) {
    *s2-- = 0;
    if (s2 == s) break;
  }
  return s;
}

char *save_fgets(char *buf, int size, FILE *f)
{
  char *res = fgets(buf,size,f);

  if (res && ( (buf[strlen(buf)-1] != '\n'))) {
    craerr(0,CRAERR_LIMIT,"'%s': Eingabezeile zu lang\n",res);
  }
  return res;
}

const char *umlaute = "";

char *strdup_umlaute(const char *ps, int flags) {
  int wieviele;
  const char *s,*s2;
  char *res;
  char *r;

  switch (flags) {
  case UMLAUTE_ISO2E: {

    wieviele = 0;
    s = ps;
    while (*s) {
      s2 = umlaute;
      while (*s2) {
	if (*s2 == *s) {
	  wieviele++;
	}
	s2++;
      }
      s++;
    }

    res = malloc(strlen(ps)+wieviele+1);

    if (!res) {
      craerr(0,CRAERR_SYS,"Nicht genug Speicher\n");
    }

    s = ps;
    r = res;
    while (*s) {
      switch (*s) {
      case '': *r++ = 'a'; *r++ = 'e'; break;
      case '': *r++ = 'o'; *r++ = 'e'; break;
      case '': *r++ = 'u'; *r++ = 'e'; break;
      case '': *r++ = 'A'; *r++ = 'e'; break;
      case '': *r++ = 'O'; *r++ = 'e'; break;
      case '': *r++ = 'U'; *r++ = 'e'; break;
      case '': *r++ = 's'; *r++ = 's'; break;
      default: *r++ = *s;
      }
      s++;
    }
    *r='\0';

    return res;
    break;
  }
  default: craerr(0,CRAERR_INTERN,"Nicht bekannte Umlautbehandlungsart: %d\n"
		  ,flags);

  /* Not reachable */
  return NULL;
  }
}

char *strassert(char* s)
{
  if (!s) {
    craerr(0,CRAERR_INPUT,"strassert");
  }
  return s;
}

char *tausenderpunkte(const char *const source, char *const target
			, size_t bufsize, char dec, char group)
{
  const char olddec = '.';
  const int groupsize = 3;

  const char *s;
  char *d;
  char *tmp = cra_alloca(bufsize);
  size_t slength = strlen(source);
  int dec_was_here; /* Gruppentrennzeichen wird nur gesetzt, wenn
		       das Dezimalzeichen schon da war oder wenn
		       gar keins vorhanden ist */
  int groupcounter;
  bool doitP; /* Bearbeitet wird nur vom Beginn einer Zahl bis
		 zu dessen Ende */

  if (slength >= bufsize) 
    craerr(0,CRAERR_INTERN
	   ,"Target buffer to small in tausenderpunkte(1) for '%s'"
	   ,source);
    
  if (!strchr(source,olddec))
    dec_was_here = T; /* Von Anfang an so tun, als ob das dezimalzeichen
			 schon dagewesen ist */
  else
    dec_was_here = NIL;

  s = source + slength-1;
  d = tmp;

  groupcounter = 0;
  doitP = NIL;
  while (s != source-1)  {
    if (! doitP && (isdigit(*s) || *s == olddec))
      doitP = T;
    if (doitP && (! (isdigit(*s) || *s == olddec)))
      doitP = NIL;
    if (*s == olddec && doitP) {
      *d = dec;
      dec_was_here = T;
    }
    else {
      if (doitP) {
	if (groupcounter == groupsize) {
	  *d++ = group;
	  groupcounter = 0;
	}
	if (dec_was_here)
	  groupcounter++;
      }
      *d = *s;
    }
    d++;
    if (d >= tmp + bufsize) 
      craerr(0,CRAERR_INTERN
	     ,"Target buffer to small in tausenderpunkte(2) for '%s'"
	     ,source);
    s--;
  } ;
  *d = 0;
  if (d >= tmp + bufsize) 
    craerr(0,CRAERR_INTERN
	   ,"Target buffer to small in tausenderpunkte(3) for '%s'"
	   ,source);

  s = tmp + strlen(tmp)-1;
  d = target;
  while (s != tmp-1) {
    *d++ = *s--;
  }
  *d = 0;

  cra_afree(tmp);
  return target;
}

char *fpformat(char *const buf, size_t bufsize
	       , const char* const fmt, ...)
{
  va_list ap;

  char *tmp = cra_alloca(bufsize);
  va_start(ap, fmt);
  vsnprintf(tmp, bufsize, fmt, ap);
  va_end(ap);

  tausenderpunkte(tmp,buf,bufsize, ',' , '.' );

  cra_afree(tmp);
  return buf;
}

void craerr(int flags, int retval, const char *fmt, ...)
{
  va_list ap;
  if (fmt) {
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    fputc('\n',stderr);
  }
  if ( flags & CRAERR_PERROR ) {
    perror("System returned");
  }

  if ( ! (flags & CRAERR_NOVERBOSE )) {

    fprintf(stderr,"Hinweis zur Fehlerbeseitigung (muss nicht exact sein):\n");

    switch (retval) {
    case CRAERR_USAGE: 
      fprintf(stderr,"Usage err - check commandline arguments\n");
      break;

    case CRAERR_INPUT:
      fprintf(stderr,"Faulty data - check the files the program works with\n");
      break;

    case CRAERR_USERSYS:
      fprintf(stderr,"Configuration error\n");
      break;

    case CRAERR_SYS:
      fprintf(stderr,"System error\n");
      break;

    case CRAERR_INTERN:
      fprintf(stderr,"Internal application check failed\n");
      break;

    case CRAERR_LIMIT:
      fprintf(stderr,"Compile-time limit exceeded\n");
      break;

    }
  }

  if (flags & CRAERR_CORE || getenv("CRAERR_CORE")) {
    abort();
  }
  if (!( flags & CRAERR_NOEXIT)) {
    fprintf(stderr,"Exiting with code %d\n",retval);
    exit(retval);
  }
  fprintf(stderr,"Mache weiter...\n\n");
}


char *bits(unsigned long wert, int laenge)
{
 static char b[sizeof(wert)*2];
 int i;
 char *p;

 p=b;
 for (i=0;i<laenge;i++) {
   if ((! (i%4)) && (i)) *p++='.';
   if (wert & ( 1<< (laenge-1-i) ))
     *p++='1';
   else
     *p++='0';
 }
 *p++=0;

 return b;
}

void
argcv(char *sp, int *argc, char *argv[], int max_argv)
{
  char *s;
  char *aktu;
  char *freep;

  aktu = malloc(strlen(sp)+2);
  freep=aktu;
  sprintf(aktu,"%s ",sp);

  *argc=0;
  while (*aktu== ' ')
    aktu++;

  while ((s=strchr(aktu,' '))) {
    while (*s== ' ')
      *s++=0;
    /* Jetzt zeigt aktu auf den Anfang eines 0-terminierten strings
       und s auf den anfang des nachsten
       */
    argv[*argc]=strdup(aktu);
    (*argc)++;
    if (*argc == max_argv) {
      fprintf(stderr,"argcv() processing of '%s' failed\n",sp);
      break;
    }
    aktu=s;
  }
  argv[*argc]=0;
  free(freep);
}


void cra_expand(char **base, char **current, char **end, size_t chunk)
{
  char *oldbase = *base;
  
  if (*current == *end) {
    
    
#ifdef HARD_DEBUG 
    if (cra_c_debug)
      fprintf(cra_c_debug,"Realloc in cra_expand %lu/%lu/%lu/%lu/%lu\n"
	      ,(unsigned long int)*base
	      ,(unsigned long int)*current
	      ,(unsigned long int)*end
	      ,(unsigned long int)chunk
	      ,(unsigned long int)oldbase
	      );
#else
    if (cra_c_debug)
      fprintf(cra_c_debug,"Realloc in cra_expand\n");
#endif

    if (! (*base = realloc(*base,*end - *base + chunk))) {
      fprintf(stderr,"Not enough memory in cra_expand\n");
      exit(2);
    }
    *end += chunk;
    *end = *end - oldbase + *base;
    *current = *current - oldbase + *base;
#ifdef HARD_DEBUG
    if (cra_c_debug)
      fprintf(cra_c_debug,"                  --> %lu/%lu/%lu/%lu/%lu\n"
	      ,(unsigned long int)*base
	      ,(unsigned long int)*current
	      ,(unsigned long int)*end
	      ,(unsigned long int)chunk
	      ,(unsigned long int)oldbase
	      );
#endif
   }
}

long cra_fsize(const char *const name)
{
  FILE *f;
  long retval;

  if (!(f = fopen(name,"r")))
    craerr(CRAERR_PERROR,CRAERR_USERSYS,"Open file '%s' failed",name);

  fseek(f,0,SEEK_END);
  retval = ftell(f);
  fclose(f);
  return retval;
}

long lfroms(const char *const source, const size_t length)
{
  char *buf = cra_alloca(length+1);
  const char *s = source;
  char *t = buf;

  while (s < (source + length)) {
    *t = *s;
    if (*s == 0)
      break; /* Zulaessig */
    s++;
    t++;
  }
  *t = 0;
  cra_afree(buf);
  return atol(buf);
}

int cra_stdtime(time_t t, char *buf, const size_t length)
{
  int n;
  struct tm *t2;
  const char *const err = "<unknown time>";
  t2 = gmtime(&t);
  if (t2 == NULL) {
    strncpy(buf, err, length);
    return 0;
  }
#if 0
  n = strftime(buf, length, "%a, %d %h %T %Y +0000", t2);
#endif
  n = strftime(buf, length, "%d %h %Y %T GMT", t2);
  if (n == 0) {
    strncpy(buf, err, length);
    return 0;
  }
  return n;
}



#ifndef HAVE_ALLOCA
void	*cra_alloca(size_t p)
{
  return malloc(p);
}
void cra_afree(void *p)
{
  free(p);
}
#endif

#ifndef HAVE_SNPRINTF
int      snprintf(char *buf, size_t bufsize, const char *fmt, ...)
{
  int retval;
  va_list ap;

  va_start(ap, fmt);
  retval = vsnprintf(buf, bufsize, fmt, ap);
  va_end(ap);
  return retval;
}
#endif
#ifndef HAVE_VSNPRINTF
int      vsnprintf(char *buf, size_t bufsize, const char *fmt, va_list va)
{
  return vsprintf(buf,fmt,va);
}
#endif
