#include <stdio.h> 
#include <stdlib.h> 
#include <stdarg.h> 
#include <string.h>

#include "cra_ps.h"

#include <cra_c.h>

#ifdef HAVE_VASPRINTF

const char *const ps_newline = "\n";

static const double gesamtscale = POINT2MM; /* kann nicht gesetzt werden
					       zuviele Stellen verlassen sich
					       auf Constanz dieses Factors.
					       Wenn man das aendern will,
					       dann sollte dies ein Parameter
					       fuer den Constructor sein
					       */

struct cra_psfont default_cra_psfonts[N_DEFAULT_CRA_PSFONTS] = { 
  { "Helvetica", 2.5 * POINT2MM, 1.0},
  { "Helvetica-Bold", 2.5 * POINT2MM, 1.0}
};


struct psoptions *C_psoptions(struct psoptions *pso)
{
  bzero(pso,sizeof(*pso));
  pso->f = stdout;
  pso->fonts.n = N_DEFAULT_CRA_PSFONTS;
  pso->fonts.fonts = malloc(sizeof(default_cra_psfonts));
  bcopy(default_cra_psfonts, pso->fonts.fonts, sizeof(default_cra_psfonts));
  return pso;
}

struct pstextopt *C_pstextopt(struct pstextopt *psto, struct psoptions *pso)
{
  bzero(psto,sizeof(psto));
  psto->pso = pso;
  psto->direction = psto_left;
  psto->font.val.i = 0; psto->font.useit = T;
  return psto;
}

void ps_fontscale(struct psoptions *pso,double factor)
{
  int i;
  for (i=0 ; i < pso->fonts.n ; i++)
    pso->fonts.fonts[i].size *= factor;
}

char *ps_expand(struct pstextopt *psto, const char *const source)
{
  const char *s;
  char *t;

  size_t length = strlen(source);
  char *target = malloc(length+1);
  char *end = target + length+1;

  s = source;
  t = target;

  while (*s) {
    if (*s == '(' || *s == ')') {
      *t = '\\'; 
      t++; cra_expand(&target,&t,&end,length);
    }

    *t = *s;

    t++; cra_expand(&target,&t,&end,length);
    s++;
  };
  *t = 0;

  return target;
}

int vpstext(struct pstextopt *psto,const char* fmt, va_list ap)
{
  struct psoptions *pso = psto->pso;
  int retval=0;

  /* Plausibilitaetschecks */
  if ( !pso->lastfont.useit && !psto->font.useit)
    craerr(0,CRAERR_INTERN,"Kein defaultfont und keiner vom User abgegeben");
  if ( psto->spaltenbreite.useit && ! psto->zeilenabstand.useit)
    craerr(0,CRAERR_INTERN,"Fuer automatischen Umbruch muss Zeilenabstand "
	   "gesetzt sein");
  if (psto->x.useit + psto->y.useit == T)
    craerr(0,CRAERR_INTERN,"Relative movement not implemented");

  /* Erst mal Font setzen */
  if (psto->font.useit) { /* User moechte, dass wir den Font setzen */
    if (psto->font.val.i != pso->lastfont.val.i) {
      /* Nur setzen, wenn nicht der gleiche wie vorher */
      fprintf(pso->f,"font%d \n",(int)psto->font.val.i);
      pso->lastfont.val.i = psto->font.val.i;
    }
  }

  /* Wohin? Es gibt diese Arten von settings:

     x.useit && y.useit : Absolute positionierung
     !x.useit && !y.useit : Wo immer der Cursor ist
     !x.useit && y.useit : not implemented (relative move)
     x.useit && !y.useit : not implemented (relative move)
     */

  if (psto->x.useit && psto->y.useit)
    fprintf(pso->f,"%g %g moveto ",psto->x.val.d,psto->y.val.d);
  else 
    if (psto->x.useit || psto->y.useit)
      craerr(0,CRAERR_INTERN,"Relative movement not implemented");

  if (psto->spaltenbreite.useit)
      craerr(0,CRAERR_INTERN,"Automatischer Umbruch not implemented");

  if (strcmp(fmt,ps_newline)) {
    fprintf(pso->f,"(");
    {
      char *s;
      char *s2;

      retval = vasprintf(&s, fmt, ap);
      s2 = ps_expand(psto,s);

      fputs(s2,pso->f);

      free(s);
      free(s2);
    }
    switch (psto->direction) {
    case psto_left:
      fprintf(pso->f,")sshow\n");
      break;
    case psto_right:
      fprintf(pso->f,")sshow_r\n");
      break;
    case psto_center:
      fprintf(pso->f,")sshow_c\n");
      break;
    case psto_right_leave:
      fprintf(pso->f,")sshow_rl\n");
      break;
    case psto_center_leave:
      fprintf(pso->f,")sshow_cl\n");
      break;
    default: 
      craerr(0,CRAERR_INTERN,"textdirection %d not known\n",psto->direction);
    }
  }

  if (psto->zeilenabstand.useit && psto->y.useit)
    psto->y.val.d -= psto->zeilenabstand.val.d;

  return retval;
}

int pstext(struct pstextopt *psto,const char* fmt, ...)
{
  int retval;

  va_list ap;
  va_start(ap, fmt);
  retval = vpstext(psto, fmt, ap);
  va_end(ap);

  return retval;
}

void ps_beginheader(const struct psoptions *const pso
		    , const char *const progname)
{
  FILE *f = pso->f;

  fprintf(f,"%%!PS-Adobe-3.0\n");
  fprintf(f,"%%%%Creator:  %s\n",progname);
  fprintf(f,"%%%%EndCommments\n");
  fprintf(f,"%%%%BeginProlog\n");
  fprintf(f,"%%%%BeginResource: procset alles\n\n");
  
  fprintf(f,"/showscaled {\n1 schriftscale scale\nshow\n");
  fprintf(f,"1 1 schriftscale div scale\n} def\n\n");
  
  fprintf(f,"/showrechtsscaled {\n"
	  "dup\n"
	  "0 exch stringwidth pop sub\n"
	  "  0 rmoveto\n");
  fprintf(f,"1 schriftscale scale\nshow\n");
  fprintf(f,"1 1 schriftscale div scale\n} def\n\n");
  
  fprintf(f,"/showcenterscaled {\n"
	  "dup\n"
	  "0 exch stringwidth pop 2 div sub\n"
	  "  0 rmoveto\n");
  fprintf(f,"1 schriftscale scale\nshow\n");
  fprintf(f,"1 1 schriftscale div scale\n} def\n\n");
  
  fprintf(f,"%%%%EndResource\n");

  /*EPS-Aktionen, Seite 733 PS-Ref.*/
  fprintf(f,
	 "/ms {moveto show } bind def\n"
	 "/s /show load def\n"
	 "/SF { %%/Fontindex FontSize /FontName SF --\n");
  fprintf(f,
	 "  findfont exch scalefont dup setfont def\n"
	 "} bind def\n"
	 "/sf /setfont load def\n"
	 "/rect { %% llx lly w h\n");
  fprintf(f,
	  " 4 2 roll moveto\n"
	  " 1 index 0 rlineto\n"
	  "  0 exch rlineto\n"
	  "  neg 0 rlineto\n"
	  "  closepath\n"
	  "} bind def\n");
  fprintf(f,
	  "/BeginEPSF {\n"
	  "/b4_Inc_state save def\n"
	  "/dict_count countdictstack def\n"
	  "/op_count count 1 sub def\n"
	  "userdict begin\n"
	  "/showpage {} def\n"
	  "0 setgray 0 setlinecap\n"
	  "1 setlinewidth 0 setlinejoin\n"
	  "10 setmiterlimit [] 0 setdash newpath\n"
	  "/languagelevel where\n"
	  "{pop laguagelevel\n"
	  "1 ne\n"
	  "{false setstrokeadjust false setoverprint\n"
	  "} if\n"
	  "} if\n"
	  "} bind def\n\n"
	  
	  "/EndEPSF {\n"
	  " count op_count sub {pop} repeat\n"
	  " countdictstack dict_count sub {end} repeat\n"
	  " b4_Inc_state restore\n"
	  " } bind def\n");
 

  fprintf(f,"%%%%EndProlog\n\n");
  fprintf(f,"%%%%BeginSetup\n");

  fprintf(f,
	  "/umlautarie {\n"
	  "/reencdict 12 dict def\n"
	  "/ReEncode {\n"
	  "reencdict begin\n\n"
	  "/newcodesandnames exch def\n"
	  "/newfontname exch def\n"
	  "/basefontname exch def\n\n"
	  "/basefontdict basefontname findfont def\n"
	  "/newfont basefontdict maxlength dict def\n\n"
	  "basefontdict {\n"
	  " exch dup /FID ne {\n"
	  "  dup /Encoding eq {\n"
	  "   exch dup length array copy newfont\n"
	  "   3 1 roll put\n"
	  "  }{\n"
	  "   exch newfont 3 1 roll put \n"
	  "  } ifelse\n"
	  " }{\n"
	  "pop pop\n"
	  " } ifelse\n"
	  "} forall\n\n"
	  "newfont /FontName newfontname put\n"
	  "newcodesandnames aload pop\n\n"
	  "newcodesandnames length 2 idiv {\n"
	  " newfont /Encoding get 3 1 roll put\n"
	  " } repeat\n\n"
	  "newfontname newfont definefont pop\n\n"
	  "end\n"
	  "} def %%ReEncode\n\n"
	  "/RZUvec [\n\n"
	  "%% dosen\n"
	  " 8#204 /adieresis\n"
	  " 8#224 /odieresis\n"
	  " 8#201 /udieresis\n"
	  " 8#216 /Adieresis\n"
	  " 8#231 /Odieresis\n"
	  " 8#232 /Udieresis\n"
	  " 8#341 /germandbls\n"
	  " \n"
	  "%% iso-8859-1/windoof\n"
	  " 8#344 /adieresis\n"
	  " 8#366 /odieresis\n"
	  " 8#374 /udieresis\n"
	  " 8#304 /Adieresis\n"
	  " 8#326 /Odieresis\n"
	  " 8#334 /Udieresis\n"
	  " 8#337 /germandbls\n"
	  " 8#351 /eacute\n"
	  " 8#264 /acute\n"
	  "\n"
	  "%% alte zaehl0-defis\n"
	  " 216 /adieresis\n"
	  " 217 /odieresis\n"
	  " 218 /udieresis\n"
	  " 219 /Adieresis\n"
	  " 220 /Odieresis\n"
	  " 221 /Udieresis\n"
	  " 222 /germandbls\n\n"
	  " 223 /copyright\n"
	  " 224 /oslash\n"
	  " 225 /eacute\n"
	  " ] def\n\n"
	  "/Helvetica /Helvetica RZUvec ReEncode\n"
	  "/Helvetica-Bold /Helvetica-Bold RZUvec ReEncode\n\n"
	  "/Times /Times RZUvec ReEncode\n"
	  "/Times-Bold /Times-Bold RZUvec ReEncode\n\n"
	  "/Courier /Courier RZUvec ReEncode\n"
	  "/Courier-Bold /Courier-Bold RZUvec ReEncode\n\n"
	  "} def 			%% umlautarie\n"
	  "umlautarie\n\n"
	  );

  fprintf(f,
	  "/fontscale 1 def\n" /* Falls zuerst normale Fonts benutzt werden */
	  "/sshow { 1 fontscale scale show 1 1 fontscale div scale} def\n"
	  "/sshow_r { dup 0 exch stringwidth pop sub \n"
	  "  0 rmoveto sshow} def\n"
	  "/sshow_rl { dup 0 exch stringwidth pop sub \n"
	  "  0 rmoveto gsave sshow grestore} def\n"
	  "/sshow_c { dup 0 exch stringwidth pop 2 div sub \n"
	  "  0 rmoveto sshow} def\n"
	  "/sshow_cl { dup 0 exch stringwidth pop 2 div sub \n"
	  "  0 rmoveto gsave sshow grestore} def\n"
	  );
  {
    int i;
    for (i=0 ; i < pso->fonts.n ; i++)
      fprintf(f,
	      "/font%d { /%s findfont %g scalefont setfont "
	      "/fontscale %g def} def\n"
	      ,i
	      ,pso->fonts.fonts[i].name
	      ,pso->fonts.fonts[i].size/gesamtscale
	      ,pso->fonts.fonts[i].vscale
	      );
  }
}

void ps_endheader(const struct psoptions *const pso)
{
  fprintf(pso->f,"%%%%EndSetup\n\n");
}

void ps_newpage(const struct psoptions *pso, int *const laufende_seitenzahl)
{

  (*laufende_seitenzahl)++;

  if (*laufende_seitenzahl!=1)
    {
      fprintf(pso->f,"pgsave restore showpage\n");
    }

  fprintf(pso->f,
	  "%%%%Page: %d %d\n"
	  "%%%%BeginPageSetup\n"
	  "/pgsave save def\n"
	  "%%%%EndPageSetup\n"
	  "%g %g scale font0\n"
	  ,*laufende_seitenzahl,*laufende_seitenzahl
	  ,gesamtscale,gesamtscale
	  );
}

void ps_endfile(const struct psoptions *const pso)
{
  fprintf(pso->f,"pgsave restore showpage\n%%EOF\n");
}
#else /* HAVE_VASPRINTF */
const char *const ps_newline = "\n";
#endif /* HAVE_VASPRINTF */
