#ifndef __CRA_UNIX_H
#define __CRA_UNIX_H

#ifdef __cplusplus
extern "C" {
#endif

  /* Resets timer mechanism */
  void timestamp(void);
  double cpu_so_far(void);
  double time_so_far(void);

  /* Return cpu and real time */
  double return_time(double *cpu, double *real);

  /* Prints time with a message to a file descriptor */
  void print_time(const char* message, FILE*);
  void print_time_stderr(void);
  void print_time_stdout(void);

  /* Signalhandler, der immer syscalls restarted */
  void cra_bsdsignal(int signo, void *alarm_catcher);

#define CRA_SETTIMER_REAL 0
#define CRA_SETTIMER_USER 1
#define CRA_SETTIMER_USERSYS 2

  void cra_settimer(int for_what_time, long sec, long usec, void (*function)(int));

  /* Returns a random number 0 <= number < max */
long cra_irand ( long max);
double cra_frand ( double max);

void unlimit(void);

/* This routine processes string s, if the pointer doesn't cause a 
   segfault or bus error. If it isn't valid, it processes
   the string `if_not_valid' instead.

   If function pointer fun is not NULL, the function will be called
   with the string to be processed. If fun is NULL, the string is just
   fputs()'ed to file f.

   */
int
cra_pifvalid(FILE *f, const char *const s, const char *const if_not_valid
	     ,void (*fun)(const char *const, FILE *f));

#ifdef __cplusplus
}
#endif


#endif
